#: Precision to which to calculate Weyl chamber coordinates or local
#: invariants.
#:
#: Making the precision too high can lead to numerical instabilities, as
#: rounding errors can put a point outside of the Weyl chamber when it should
#: be inside, or into a different region.
DEFAULT_WEYL_PRECISSION = 8
